// From the software distribution accompanying the textbook
// "A Practical Introduction to Data Structures and Algorithm Analysis,
// Third Edition (C++)" by Clifford A. Shaffer.
// Source code Copyright (C) 2007-2011 by Clifford A. Shaffer.

// This file includes all of the pieces of the BST implementation

// Include the node implementation
#include "binnode.h"

// Include the dictionary ADT
#include "dictionary.h"

// Binary Search Tree implementation for the Dictionary ADT
template <typename Key, typename Elem>
class BST : public Dictionary<Key,Elem> {
private:
  BNode<Key,Elem>* root;   // Root of the BST
  int nodecount;         // Number of nodes in the BST

  // Private "helper" functions
  void clearhelp(BNode<Key, Elem>*);
  BNode<Key,Elem>* inserthelp(BNode<Key, Elem>*,
                              const Key&, const Elem&);
  BNode<Key,Elem>* deletemin(BNode<Key, Elem>*,
                             BNode<Key, Elem>*&);
  BNode<Key,Elem>* removehelp(BNode<Key, Elem>*, const Key&,
                              BNode<Key, Elem>*&);
  bool findhelp(BNode<Key, Elem>*, const Key&, Elem&) const;
  void printhelp(BNode<Key, Elem>*, int) const;

public:
  BST() { root = NULL; nodecount = 0; }  // Constructor
  ~BST() { clearhelp(root); }            // Destructor

  void clear()   // Reinitialize dictionary
    { clearhelp(root); root = NULL; nodecount = 0; }

  // Insert Elem "it" with Key "k" into the dictionary.
  void insert(const Key& k, const Elem& it) {
    root = inserthelp(root, k, it);
    nodecount++;
  }

  // Remove some record with key "K" and return it in "it".
  // Return true if such exists, false otherwise.
  // If multiple records match "K", remove an arbitrary one.
  bool remove(const Key& K, Elem& it) {
    BNode<Key,Elem>* t = NULL;
    root = removehelp(root, K, t);
    if (t == NULL) return false;  // Nothing found
    it = t->val();
    nodecount--;
    delete t;
    return true;
  }
  // Remove and return the record with smallest key value
  // from the dictionary and return in "it". Return true if
  // some record is removed, false otherwise.
  Elem removeAny() {  // Delete min value
    Assert(root != NULL, "Empty tree");
    BNode<Key, Elem>* t;
    root = deletemin(root, t);
    Elem it = t->val();
    delete t;
    nodecount--;
    return it;
  }

  // Return a copy in "it" of some record matching "K".
  // Return true if such exists, false otherwise. If
  // multiple records match "K", return an arbitrary one.
  bool find(const Key& K, Elem& it) const
    { return findhelp(root, K, it); }

  // Return the number of records in the dictionary.
  int size() { return nodecount; }

  void print() const { // Print the contents of the BST
    if (root == NULL) cout << "The BST is empty.\n";
    else printhelp(root, 0);
  }
};

// Clean up BST by releasing space back free store
template <typename Key, typename Elem>
void BST<Key, Elem>::
clearhelp(BNode<Key, Elem>* root) {
  if (root == NULL) return;
  clearhelp(root->left());
  clearhelp(root->right());
  delete root;
}

// Insert a node into the BST, returning the updated tree
template <typename Key, typename Elem>
BNode<Key, Elem>* BST<Key, Elem>::inserthelp(
    BNode<Key, Elem>* root, const Key& k, const Elem& it) {
  if (root == NULL)  // Empty tree: create node
    return new BNode<Key, Elem>(k, it, NULL, NULL);
  if (k < root->key())
    root->setLeft(inserthelp(root->left(), k, it));
  else root->setRight(inserthelp(root->right(), k, it));
  return root;       // Return tree with node inserted
}

// Delete the minimum value from the BST, returning the revised BST
template <typename Key, typename Elem>
BNode<Key, Elem>* BST<Key, Elem>::
deletemin(BNode<Key, Elem>* root, BNode<Key, Elem>*& S) {
  if (root->left() == NULL) { // Found min
    S = root;
    return root->right();
  }
  else {                      // Continue left
    root->setLeft(deletemin(root->left(), S));
    return root;
  }
}

// Return in "R" the element (if any) with value "K"
// Return the updated subtree with "R" removed from the tree
template <typename Key, typename Elem>
BNode<Key, Elem>* BST<Key, Elem>::
removehelp(BNode<Key, Elem>* root, const Key& K,
           BNode<Key, Elem>*& R) {
  if (root == NULL) return NULL;    // K is not in tree
  else if (K < root->key())
    root->setLeft(removehelp(root->left(), K, R));
  else if (K > root->key())
    root->setRight(removehelp(root->right(), K, R));
  else {                            // Found: remove it
    BNode<Key, Elem>* temp;
    R = root;
    if (root->left() == NULL)       // Only a right child
      root = root->right();         //  so point to right
    else if (root->right() == NULL) // Only a left child
      root = root->left();          //  so point to left
    else {                    // Both children are non-empty
      root->setRight(deletemin(root->right(), temp));
      Elem te = root->val();
      root->setVal(temp->val());
      temp->setVal(te);
      R = temp;
    }
  }
  return root;
}

// Find a node with the given key value
template <typename Key, typename Elem>
bool BST<Key, Elem>::findhelp(BNode<Key, Elem>* root,
                              const Key& K, Elem& e) const {
  if (root == NULL) return false;          // Empty tree
  else if (K < root->key())
    return findhelp(root->left(), K, e);   // Check left
  else if (K > root->key())
    return findhelp(root->right(), K, e);  // Check right
  else { e = root->val();  return true; }  // Found it
}

// Print out a BST
template <typename Key, typename Elem>
void BST<Key, Elem>::
printhelp(BNode<Key, Elem>* root, int level) const {
  if (root == NULL) return;           // Empty tree
  printhelp(root->left(), level+1);   // Do left subtree
  for (int i=0; i<level; i++)         // Indent to level
    cout << "  ";
  cout << root->key() << "\n";        // Print node value
  printhelp(root->right(), level+1);  // Do right subtree
}
